/***************************************************************************************
	Fischertechnik Universal Interface 30520 als Porterweiterung am Atmel AVR.

*** INKLUSIVE PWM MOTORSTEUERUNG *** 
*** Variante mit Nutzung eines einzigen Interrupts (Timer 2 Overflow)***

	Weitere Hinweise siehe FTUI_PWM.h

	Es wird keine Haftung bernommen! Die Benutzung dieser Routine geschieht
	auf eigene Gefahr!

	Autor:	Dirk Uffmann, mailto: uffmann@arcor.de

	Datum:	27.11.2015

***************************************************************************************/

#include "FTUI_PWM.h"

#if slave
	volatile uint16_t motorword;
	volatile uint16_t motorcontrol;
	volatile uint16_t inputword;
#else
	volatile uint8_t motorword;
	volatile uint8_t motorcontrol;
	volatile uint8_t inputword;
#endif

	volatile uint8_t pwm_table[4 + (4*slave)];
	volatile uint8_t pwm_counter = 0;

/******************************************************************************
	FTUI_update bertrgt das motorword und liest parallel das inputword ein.
	Diese Funktion muss mindestens alle 300 ms aufgerufen werden, um den 
	Watchdog im FT-Interface zurckzusetzen (durch das Clocksignal). Das 
	passiert hier automatisch ber den Timer2-Overflow-Interrupt mindestens 
	alle 12,8 ms (100 * 128s).
	Bei einem Systemtakt von 16 MHz braucht die Routine mit einem Interface
	nur ca. 12 s, mit zwei Interfaces sind es ca. 32 s.
******************************************************************************/
inline void FTUI_update (void)
{
#if slave
	uint16_t mask_motor=(1<<15);
	uint16_t mask_input=(1<<7);
#else
	uint8_t mask_motor=(1<<7);
	uint8_t mask_input=(1<<7);
#endif
//	Clock low-Phase einleiten
	FTUI_Port_Write &= ~(1<<FTUI_Clock); 

//	Zunchst werden alle Eingnge als 0 angenommen.
	inputword = 0;

//	CMOS4014: Nun werden die Eingangszustnde in das Schieberegister bernommen.
//	Dazu mu LoadIn gesetzt werden und danach eine positive Clockflanke erzeugt werden.
	FTUI_Port_Write |= (1<<FTUI_LoadIn);
	FTUI_Port_Write |= (1<<FTUI_Clock); //Clock high-Phase einleiten
//	mit dieser positiven Clockflanke steht das erste Bit auf der
//	DataIn-Leitung an, jedoch mit einer Verzgerung. Daher wird
//	es erst in der nchsten Clock-Low-Phase eingelesen

//	CMOS4014: Dann werden die Eingnge wieder vom Schieberegister entkoppelt
	FTUI_Port_Write &= ~(1<<FTUI_LoadIn);

//	Dann werden die Bits 7 bis 0 (ein FT-Interface) bzw. 15 bis 0 (zwei
//	FT-Interfaces) gesendet (motorword) und parallel die Bits 7 bis 0 und
//	dann 15 bis 8 (zwei FT-Interfaces) eingelesen (inputword)
	while (mask_motor)
	{
//		Wenn das jeweilige Sende-Bit high sein soll, dann wird DataOut = 1
//		gesetzt, sonst DataOut = 0
		if(motorword & mask_motor) FTUI_Port_Write |= (1<<FTUI_DataOut);
		else FTUI_Port_Write &= ~(1<<FTUI_DataOut);
		FTUI_Port_Write &= ~(1<<FTUI_Clock); //Clock low-Phase einleiten
		mask_motor = (mask_motor>>1); // Schiebe das Maskierungsbit eins nach rechts

//		Nun wird das jeweilige Bit im Datenstream auf der Leitung DataIn
//		daraufhin getestet, ob der zugehrige Eingang "high" ist,
//		und dann wird inputword an der jeweilige Bitposition auf 1 korrigiert
		if (FTUI_Port_Read & (1<<FTUI_DataIn)) inputword |= mask_input;
		mask_input = (mask_input>>1); //Maskierungsbit nach rechts weiterschieben
#if slave
		if (mask_input==0) mask_input = (1<<15); //Maskierungsbit auf Position 15 setzen
#endif
		FTUI_Port_Write |= (1<<FTUI_Clock); //Clock high-Phase einleiten
	}

//	CMOS4094: Durchschalten der Daten aus dem Schieberegister ins Speicherregister
//	fr die Motorausgnge, dieses Strobe-Signal muss 200ns andauern (> 3 Takte 16MHz)
	FTUI_Port_Write |= (1<<FTUI_LoadOut);
	FTUI_Port_Write |= (1<<FTUI_LoadOut);
	FTUI_Port_Write |= (1<<FTUI_LoadOut);

//	wieder Entkoppeln von Speicherregister und Schieberegister fr die Ausgnge
	FTUI_Port_Write &= ~(1<<FTUI_LoadOut);
}

ISR (TIMER2_OVF_vect) //findet alle 128s statt
{
	uint8_t i;
	uint8_t update_flag = 0;
	pwm_counter++;
	if (pwm_counter == 100)
	{
		pwm_counter = 0;
//	Ausgnge gem der Variable motorcontrol einschalten
		motorword = motorcontrol;
		update_flag = 1;
	}

	i = 4 + (4*slave);
	do 
	{
		i--;
		if (pwm_table[i] == pwm_counter)
		{
//	Ausgnge gem der PWM Tabelle pwm_table[] abschalten
			if (output & (0x0003 << (2*i))) motorword |= (0x0003 << (2*i));
			else motorword &= ~(0x0003 << (2*i));
			update_flag = 1;
		}
	} while (i);
	
	if (update_flag) FTUI_update();
}

/****************************************************************************
	FTUI Init Routine
****************************************************************************/
void FTUI_Init (void)
{
//	internen Pull-up Widerstand fr DataIn einschalten
	FTUI_Port_Write = (1<<FTUI_DataIn);
//	Ausgnge auf low setzen
	FTUI_Port_Write &= ~((1<<FTUI_LoadIn)|(1<<FTUI_DataOut)|(1<<FTUI_Clock)|(1<<FTUI_LoadOut));
//	Ausgnge definieren bzw. aktiv schalten
	FTUI_Port_DDR = (1<<FTUI_LoadOut)|(1<<FTUI_LoadIn)|(1<<FTUI_DataOut)|(1<<FTUI_Clock);
//	system clock 16 MHz is used (external quartz oscillator)
//	start timer2 with prescaler clock/8
	TCCR2B = _BV(CS21);
//	activate 8-bit Timer0 overflow interrupt with prescaler divide by 8 -> 8 * 256 / (16 MHz) = 128 s or ca. 7,8 kHz
	TIMSK2 = _BV(TOIE2);
	sei();

//	Motorausgnge initial high oder low schalten entsprechend der gewhlten Beschaltung
#if slave
	motorcontrol = output;
#else
	motorcontrol = output1;
#endif
}

/*********************************************************************************
	M () stellt die Funktion zur Manipulation von motorcontrol zur Steuerung 
	einzelner Motoren bereit. Der pwm Wert gilt auch bei Einzelbeschaltung
	der Ausgnge immer fr beide Ausgnge.
*********************************************************************************/
void M (uint8_t motor_number, uint8_t state, uint8_t pwm)
{
//	Abfangen unsinniger Werte
	if (motor_number == 0) return;
	if (motor_number > (4 + (4*slave))) return;
	if (state > 3) return;

	uint8_t position = (motor_number-1)*2;
	pwm_table[motor_number-1] = pwm;

//	Zunchst wird "state" in "motorcontrol" an der gewnschten Position hinzugefuegt
//	dazu werden zunchst die jeweiligen 2 bits an den Bit-Positionen fr motor_number
//	auf 0 gesetzt
	motorcontrol &= ~(3 << position);
//	und dann die 2 neuen Bits dem motorcontrol hinzugefgt
	motorcontrol |= state << position;
}

/********************************************************************************
	E () stellt die Funktion zum Dekodieren einzelner Eingnge aus dem inputword
	bereit. Zuvor muss inputword mittels FTUI_update() eingelesen worden sein.
	input_number ist die Eingangsnummer (1-8 bzw 1-16 mit zustzlichem Slave-
	interface am Erweiterungsport) die dekodiert werden soll.
********************************************************************************/
uint8_t E (uint8_t input_number)
{
	uint8_t input;

//	Wenn das inputword an der Bitposition von 
//	"input_number" > 0 dann gib 1 zurck, ansonsten 0
	if (inputword & (1<<(input_number-1))) input = 1;
	else input = 0;

//	Rckgabewert
	return input;
}

